import staircase as sc

from piso._exceptions import ClosedValueError, DegenerateIntervalError


def _validate_intervals(interval_array):
    if not all(interval_array.length):  # test for degenerate intervals
        raise DegenerateIntervalError(interval_array)
    if interval_array.closed not in ("left", "right"):
        raise ClosedValueError


def _interval_x_to_stairs(interval_array):
    # can be used with interval, interval array, interval index
    assert interval_array.closed in {"left", "right"}
    return sc.Stairs(
        start=interval_array.left,
        end=interval_array.right,
        closed=interval_array.closed,
    )


def _boolean_stairs_to_interval_array(stairs, cls):
    if stairs.identical(0):
        return cls([], closed=stairs.closed)
    return cls.from_arrays(
        stairs.step_changes.index[::2],
        stairs.step_changes.index[1::2],
        closed=stairs.closed,
    )



from piso.intervalarray import difference, intersection, symmetric_difference, union


def register_accessors():
    """
    When called this function will register the "piso" ArrayAccessor on :class:`pandas.IntervalIndex` and :class:`pandas.arrays.IntervalArray`.

    Examples
    --------

    >>> import piso
    >>> piso.register_accessors()
    >>> arr = pd.arrays.IntervalArray.from_tuples(
    ...     [(0, 4), (2, 5), (3, 6), (7, 8), (8, 9), (10, 12)],
    ... )

    >>> arr.piso.union()
    <IntervalArray>
    [(0.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
    Length: 3, closed: right, dtype: interval[float64]

    >>> arr = pd.IntervalIndex(arr)
    >>> arr.piso.union()
    IntervalIndex([(0.0, 2.0], (5.0, 6.0], (7.0, 9.0], (10.0, 12.0]],
                closed='right',
                dtype='interval[float64]')
    """
    from piso import accessor

    accessor._register_accessors()


def get_version():
    def get_version_post_py38():
        from importlib.metadata import version  # type: ignore

        return version(__name__)

    def get_version_pre_py38():
        from pkg_resources import get_distribution

        return get_distribution(__name__).version

    def default_version():
        return "unknown"

    for func in (get_version_post_py38, get_version_pre_py38, default_version):
        try:
            return func()
        except Exception:
            pass


__version__ = get_version()



union_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso.interval

>>> piso.interval.union(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
... )
Interval(0.0, 4.0, closed='right')

>>> piso.interval.union(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
...    squeeze=False,
... )
<IntervalArray>
[(0.0, 4.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> piso.interval.union(
...    pd.Interval(0, 3, closed="left"),
...    pd.Interval(2, 4, closed="left"),
... )
Interval(0.0, 4.0, closed='left')

>>> piso.interval.union(
...    pd.Interval(0, 1),
...    pd.Interval(3, 4),
... )
<IntervalArray>
[(0.0, 1.0], (3.0, 4.0]]
Length: 2, closed: right, dtype: interval[float64]
"""

intersection_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso.interval

>>> piso.interval.intersection(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
... )
Interval(2.0, 3.0, closed='right')

>>> piso.interval.intersection(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
...    squeeze=False,
... )
<IntervalArray>
[(2.0, 3.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> piso.interval.intersection(
...    pd.Interval(0, 3, closed="left"),
...    pd.Interval(2, 4, closed="left"),
... )
Interval(2.0, 3.0, closed='left')

>>> piso.interval.intersection(
...    pd.Interval(0, 1),
...    pd.Interval(3, 4),
... )
<IntervalArray>
[]
Length: 0, closed: right, dtype: interval[int64]
"""

difference_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso.interval

>>> piso.interval.difference(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
... )
Interval(0.0, 2.0, closed='right')

>>> piso.interval.difference(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
...    squeeze=False,
... )
<IntervalArray>
[(0.0, 2.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> piso.interval.difference(
...    pd.Interval(0, 4, closed="left"),
...    pd.Interval(2, 3, closed="left"),
... )
<IntervalArray>
[[0.0, 2.0), [3.0, 4.0)]
Length: 2, closed: left, dtype: interval[float64]

>>> piso.interval.difference(
...    pd.Interval(2, 3),
...    pd.Interval(0, 4),
... )
<IntervalArray>
[]
Length: 0, closed: right, dtype: interval[int64]
"""


symmetric_difference_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso.interval

>>> piso.interval.symmetric_difference(
...    pd.Interval(0, 3),
...    pd.Interval(2, 4),
... )
<IntervalArray>
[(0.0, 2.0], (3.0, 4.0]]
Length: 2, closed: right, dtype: interval[float64]

>>> piso.interval.symmetric_difference(
...    pd.Interval(0, 3),
...    pd.Interval(2, 3),
... )
Interval(0.0, 2.0, closed='right')

>>> piso.interval.symmetric_difference(
...    pd.Interval(0, 3, closed="left"),
...    pd.Interval(2, 4, closed="left"),
... )
<IntervalArray>
[[0.0, 2.0), [3.0, 4.0)]
Length: 2, closed: left, dtype: interval[float64]

>>> piso.interval.symmetric_difference(
...    pd.Interval(2, 3),
...    pd.Interval(0, 4),
... )
<IntervalArray>
[(0.0, 2.0], (3.0, 4.0]]
Length: 2, closed: right, dtype: interval[float64]
"""

template_doc = """
Performs the {operation} of two pandas.Intervals

Parameters
----------
interval1 : pandas.Interval
    the first operand
interval2 : pandas.Interval
    the second operand
squeeze : boolean, default True
    If True, will try to coerce the return value to a pandas.Interval

Returns
----------
:class:`pandas.Interval` or :class:`pandas.arrays.IntervalArray`

{examples}
"""

union_docstring = template_doc.format(operation="union", examples=union_examples)
intersection_docstring = template_doc.format(
    operation="intersection", examples=intersection_examples
)
difference_docstring = template_doc.format(
    operation="set difference", examples=difference_examples
)
symmetric_difference_docstring = template_doc.format(
    operation="symmetric difference", examples=symmetric_difference_examples
)



from textwrap import dedent
from typing import Any, Callable, Optional, TypeVar

FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)


class Appender:

    """
    A function decorator that will append an addendum to the docstring
    of the target function.

    This decorator should be robust even if func.__doc__ is None
    (for example, if -OO was passed to the interpreter).

    Usage: construct a docstring.Appender with a string to be joined to
    the original docstring. An optional 'join' parameter may be supplied
    which will be used to join the docstring and addendum. e.g.

    add_copyright = Appender("Copyright (c) 2009", join='\n')

    @add_copyright
    def my_dog(has='fleas'):
        "This docstring will have a copyright below"
        pass

    Notes
    -----
    The definition for this method has been borrowed, almost verbatim, from :mod:`pandas`
    """

    addendum: Optional[str]

    def __init__(self, addendum: Optional[str], join: str = "", indents: int = 0):
        if indents > 0:
            self.addendum = indent(addendum, indents=indents)
        else:
            self.addendum = addendum
        self.join = join

    def __call__(self, func: F) -> F:
        func.__doc__ = func.__doc__ if func.__doc__ else ""
        self.addendum = self.addendum if self.addendum else ""
        docitems = [func.__doc__, self.addendum]
        func.__doc__ = dedent(self.join.join(docitems))
        return func


def indent(text: Optional[str], indents: int = 1) -> str:
    if not text or not isinstance(text, str):
        return ""
    jointext = "".join(["\n"] + ["    "] * indents)
    return jointext.join(text.split("\n"))



import pandas as pd
import staircase as sc

import piso.docstrings.intervalarray as docstrings
from piso._decorators import Appender
from piso.util import (
    _boolean_stairs_to_interval_array,
    _interval_x_to_stairs,
    _validate_intervals,
)


def _check_matched_closed(interval_arrays):
    closed_values = [arr.closed for arr in interval_arrays]
    assert closed_values.count(closed_values[0]) == len(closed_values)


def _validate_array_of_intervals_arrays(*interval_arrays):
    assert len(interval_arrays) > 0
    _check_matched_closed(interval_arrays)
    for arr in interval_arrays:
        _validate_intervals(arr)


def _get_return_type(interval_array, return_type):
    assert return_type in ("infer", pd.IntervalIndex, pd.arrays.IntervalArray)
    return interval_array.__class__ if return_type == "infer" else return_type


def _make_stairs(*interval_arrays):
    if len(interval_arrays) == 1:
        stairs = _interval_x_to_stairs(*interval_arrays)
    else:
        stairs = sc.sum(
            [_interval_x_to_stairs(arr).make_boolean() for arr in interval_arrays]
        )
    return stairs


@Appender(docstrings.union_docstring, join="\n", indents=1)
def union(interval_array, *interval_arrays, squeeze=False, return_type="infer"):
    _validate_array_of_intervals_arrays(interval_array, *interval_arrays)
    klass = _get_return_type(interval_array, return_type)
    stairs = _make_stairs(interval_array, *interval_arrays)
    result = _boolean_stairs_to_interval_array(stairs.make_boolean(), klass)
    if squeeze and len(result) == 1:
        result = result[0]
    return result


@Appender(docstrings.intersection_docstring, join="\n", indents=1)
def intersection(
    interval_array,
    *interval_arrays,
    min_overlaps="all",
    squeeze=False,
    return_type="infer"
):
    _validate_array_of_intervals_arrays(interval_array, *interval_arrays)
    klass = _get_return_type(interval_array, return_type)
    if min_overlaps == "all":
        min_overlaps = (
            len(interval_arrays) + 1 if interval_arrays else len(interval_array)
        )
    stairs = _make_stairs(interval_array, *interval_arrays)
    result = _boolean_stairs_to_interval_array(stairs >= min_overlaps, klass)
    if squeeze and len(result) == 1:
        result = result[0]
    return result


@Appender(docstrings.difference_docstring, join="\n", indents=1)
def difference(interval_array, *interval_arrays, squeeze=False, return_type="infer"):
    assert interval_arrays
    _validate_array_of_intervals_arrays(interval_array, *interval_arrays)
    klass = _get_return_type(interval_array, return_type)
    stairs_operand1 = _interval_x_to_stairs(interval_array)
    stairs_operand2 = _make_stairs(*interval_arrays)
    stairs = stairs_operand1 & (~stairs_operand2)
    result = _boolean_stairs_to_interval_array(stairs, klass)
    if squeeze and len(result) == 1:
        result = result[0]
    return result


@Appender(docstrings.symmetric_difference_docstring, join="\n", indents=1)
def symmetric_difference(
    interval_array, *interval_arrays, min_overlaps=2, squeeze=False, return_type="infer"
):
    _validate_array_of_intervals_arrays(interval_array, *interval_arrays)
    klass = _get_return_type(interval_array, return_type)
    if min_overlaps == "all":
        min_overlaps = (
            len(interval_arrays) + 1 if interval_arrays else len(interval_array)
        )
    stairs = _make_stairs(interval_array, *interval_arrays)

    if min_overlaps == 2:
        stairs = stairs == 1
    else:
        stairs = (stairs >= 1) & (stairs <= min_overlaps - 1)
    result = _boolean_stairs_to_interval_array(stairs, klass)
    if squeeze and len(result) == 1:
        result = result[0]
    return result



import warnings

import pandas as pd

import piso.docstrings.accessor as docstrings
from piso import intervalarray
from piso._decorators import Appender


class CachedAccessor:
    """
    Custom property-like object.
    A descriptor for caching accessors.

    Parameters
    ----------
    name : str
        Namespace that will be accessed under, e.g. ``df.foo``.
    accessor : cls
        Class with the extension methods.

    Notes
    -----
    The definition for CachedAccessor has been borrowed, almost verbatim, from :mod:`pandas`
    """

    def __init__(self, name: str, accessor) -> None:
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., IntervalIndex.piso
            return self._accessor
        accessor_obj = self._accessor(obj)
        # Replace the property with the accessor object. Inspired by:
        # https://www.pydanny.com/cached-property.html
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


def _register_accessor(name, cls):
    """
    Notes
    -----
    The definition for this method has been borrowed, almost verbatim, from :mod:`pandas`
    """

    def decorator(accessor):
        if hasattr(cls, name) and getattr(cls, name) != ArrayAccessor:
            warnings.warn(
                f"registration of accessor {repr(accessor)} under name "
                f"{repr(name)} for type {repr(cls)} is overriding a preexisting "
                f"attribute with the same name.",
                UserWarning,
                stacklevel=2,
            )
        setattr(cls, name, CachedAccessor(name, accessor))
        if not hasattr(cls, "_accessors"):
            cls._accessors = set()
        cls._accessors.add(name)
        return accessor

    return decorator


class ArrayAccessor:
    def __init__(self, _interval_array):
        self._interval_array = _interval_array

    @Appender(docstrings.union_docstring, join="\n", indents=1)
    def union(self, *interval_arrays, squeeze=False, return_type="infer"):
        return intervalarray.union(
            self._interval_array,
            *interval_arrays,
            squeeze=squeeze,
            return_type=return_type,
        )

    @Appender(docstrings.intersection_docstring, join="\n", indents=1)
    def intersection(
        self, *interval_arrays, min_overlaps="all", squeeze=False, return_type="infer"
    ):
        return intervalarray.intersection(
            self._interval_array,
            *interval_arrays,
            min_overlaps=min_overlaps,
            squeeze=squeeze,
            return_type=return_type,
        )

    @Appender(docstrings.difference_docstring, join="\n", indents=1)
    def difference(self, *interval_arrays, squeeze=False, return_type="infer"):
        return intervalarray.difference(
            self._interval_array,
            *interval_arrays,
            squeeze=squeeze,
            return_type=return_type,
        )

    @Appender(docstrings.symmetric_difference_docstring, join="\n", indents=1)
    def symmetric_difference(
        self, *interval_arrays, min_overlaps=2, squeeze=False, return_type="infer"
    ):
        return intervalarray.symmetric_difference(
            self._interval_array,
            *interval_arrays,
            min_overlaps=min_overlaps,
            squeeze=squeeze,
            return_type=return_type,
        )


def _register_accessors():
    _register_accessor("piso", pd.IntervalIndex)(ArrayAccessor)
    _register_accessor("piso", pd.arrays.IntervalArray)(ArrayAccessor)



union_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso
>>> piso.register_accessors()


Examples with *interval_arrays* empty:

>>> arr = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (2, 5), (3, 6), (7, 8), (8, 9), (10, 12)],
... )

>>> arr.piso.union()
<IntervalArray>
[(0.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
Length: 3, closed: right, dtype: interval[float64]

>>> arr.set_closed("left").piso.union()
<IntervalArray>
[[0, 4), [2, 5), [3, 6), [7, 8), [8, 9), [10, 12)]
Length: 6, closed: left, dtype: interval[int64]

>>> pd.IntervalIndex(arr).piso.union()
IntervalIndex([(0.0, 6.0], (7.0, 9.0], (10.0, 12.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr.piso.union(return_type=pd.IntervalIndex)
IntervalIndex([(0.0, 6.0], (7.0, 9.0], (10.0, 12.0]],
              closed='right',
              dtype='interval[float64]')


Examples with *interval_arrays* non empty:

>>> arr1 = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (5, 6), (7, 8), (10, 12)],
... )
>>> arr2 = pd.arrays.IntervalArray.from_tuples(
...     [(3, 5), (8, 9)],
... )
>>> arr3 = pd.arrays.IntervalArray.from_tuples(
...     [(6, 8), (9, 11)],
... )

>>> arr1.piso.union(arr2)
<IntervalArray>
[(0.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
Length: 3, closed: right, dtype: interval[float64]

>>> arr2.piso.union(arr3, return_type=pd.IntervalIndex)
IntervalIndex([(3.0, 5.0], (6.0, 11.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr1.piso.union(arr2, arr3)
<IntervalArray>
[(0.0, 12.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> arr1.piso.union(arr2, arr3, squeeze=True)
Interval(0.0, 12.0, closed='right')
"""

intersection_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso
>>> piso.register_accessors()


Examples with *interval_arrays* empty:

>>> arr = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (2, 5), (3, 6)],
... )

>>> arr.piso.intersection()
<IntervalArray>
[(3.0, 4.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> pd.IntervalIndex(arr).piso.intersection()
IntervalIndex([(3.0, 4.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr.piso.intersection(return_type=pd.IntervalIndex)
IntervalIndex([(3.0, 4.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr.piso.intersection(min_overlaps=2)
<IntervalArray>
[(2.0, 5.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> arr2 = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (2, 5), (3, 6), (7, 8), (8, 9), (10, 12)],
... )

>>> arr.piso.intersection()
<IntervalArray>
[]
Length: 0, closed: right, dtype: interval[int64]

>>> arr.piso.intersection(min_overlaps=2)
<IntervalArray>
[(2.0, 5.0]]
Length: 1, closed: right, dtype: interval[float64]

Examples with *interval_arrays* not empty:

>>> arr1 = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (5, 6), (7, 8), (10, 12)],
... )
>>> arr2 = pd.arrays.IntervalArray.from_tuples(
...     [(3, 5), (8, 9)],
... )
>>> arr3 = pd.arrays.IntervalArray.from_tuples(
...     [(6, 8), (9, 11)],
... )

>>> arr1.piso.intersection(arr2)
<IntervalArray>
[(3.0, 4.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> arr1.piso.intersection(arr2, squeeze=True)
Interval(3.0, 4.0, closed='right')

>>> arr1.piso.intersection(arr2, arr3)
<IntervalArray>
[]
Length: 0, closed: right, dtype: interval[float64]

>>> arr1.piso.intersection(arr2, arr3, min_overlaps=2)
<IntervalArray>
[(3.0, 4.0], (10.0, 11.0]]
Length: 2, closed: right, dtype: interval[float64]
"""

difference_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso
>>> piso.register_accessors()


>>> arr1 = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (2, 5), (3, 6), (7, 8), (8, 9), (10, 12)],
... )
>>> arr2 = pd.arrays.IntervalArray.from_tuples(
...     [(4, 7), (8, 11)],
... )
>>> arr3 = pd.arrays.IntervalArray.from_tuples(
...     [(2, 5), (7, 13)],
... )

>>> arr1.piso.difference(arr2)
<IntervalArray>
[(0.0, 4.0], (7.0, 8.0], (11.0, 12.0]]
Length: 3, closed: right, dtype: interval[float64]

>>> arr1.set_closed("left").piso.difference(arr2.set_closed("left"))
<IntervalArray>
[[0.0, 4.0), [7.0, 8.0), [11.0, 12.0)]
Length: 3, closed: left, dtype: interval[float64]

>>> arr1.piso.difference(arr2, return_type=pd.IntervalIndex)
IntervalIndex([(0.0, 4.0], (7.0, 8.0], (11.0, 12.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr1.piso.difference(arr2, arr3)
<IntervalArray>
[(0.0, 2.0]]
Length: 1, closed: right, dtype: interval[float64]

>>> arr1.piso.difference(arr2, arr3, squeeze=True)
Interval(0.0, 2.0, closed='right')
"""

symmetric_difference_examples = """
Examples
-----------

>>> import pandas as pd
>>> import piso
>>> piso.register_accessors()


Examples with *interval_arrays* empty:

>>> arr = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (2, 5), (3, 6), (7, 8), (8, 9), (10, 12)],
... )

>>> arr.piso.symmetric_difference()
<IntervalArray>
[(0.0, 2.0], (5.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
Length: 4, closed: right, dtype: interval[float64]

>>> pd.IntervalIndex(arr).piso.symmetric_difference()
IntervalIndex([(0.0, 2.0], (5.0, 6.0], (7.0, 9.0], (10.0, 12.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr.piso.symmetric_difference(return_type=pd.IntervalIndex)
IntervalIndex([(0.0, 2.0], (5.0, 6.0], (7.0, 9.0], (10.0, 12.0]],
              closed='right',
              dtype='interval[float64]')

>>> arr.piso.symmetric_difference(min_overlaps=3)
<IntervalArray>
[(0.0, 3.0], (4.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
Length: 4, closed: right, dtype: interval[float64]

>>> arr.piso.symmetric_difference(min_overlaps="all")
<IntervalArray>
[(0.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
Length: 3, closed: right, dtype: interval[float64]

Examples with *interval_arrays* non-empty:

>>> arr1 = pd.arrays.IntervalArray.from_tuples(
...     [(0, 4), (5, 6), (7, 8), (10, 12)],
... )
>>> arr2 = pd.arrays.IntervalArray.from_tuples(
...     [(3, 5), (8, 9)],
... )
>>> arr3 = pd.arrays.IntervalArray.from_tuples(
...     [(6, 8), (9, 11)],
... )

>>> arr1.piso.symmetric_difference(arr2)
<IntervalArray>
[(0.0, 3.0], (4.0, 6.0], (7.0, 9.0], (10.0, 12.0]]
Length: 4, closed: right, dtype: interval[float64]

>>> arr1.piso.symmetric_difference(arr2, arr3)
<IntervalArray>
[(0.0, 3.0], (4.0, 7.0], (8.0, 10.0], (11.0, 12.0]]
Length: 4, closed: right, dtype: interval[float64]

>>> arr1.piso.symmetric_difference(arr2, arr3, min_overlaps="all")
<IntervalArray>
[(0.0, 12.0]]
Length: 1, closed: right, dtype: interval[float64]
"""


def join_params(list_of_param_strings):
    return "".join(list_of_param_strings).replace("\n\n", "\n")


param_optional_args = """
*interval_arrays : argument list of :class:`pandas.IntervalIndex` or :class:`pandas.arrays.IntervalArray`
    May contain zero or more arguments.
"""

param_optional_args_difference = """
*interval_arrays : argument list of :class:`pandas.IntervalIndex` or :class:`pandas.arrays.IntervalArray`
    Must contain at least one argument.
"""

param_min_overlaps = """
min_overlaps : int or "all", default "all"
    Specifies the minimum number of intervals which overlap in order to define an *intersection*.
    If *min_overlaps* is an int then it must be no smaller than 2.  If *min_overlaps* is all then
    an intersection is only defined where every interval overlaps.
    If supplied, must be done so as a keyword argument.
"""

param_squeeze = """
squeeze : boolean, default True
    If True, will try to coerce the return value to a pandas.Interval.
    If supplied, must be done so as a keyword argument.
"""

param_return_type = """
return_type : {"infer", :class:`pandas.IntervalIndex`, :class:`pandas.arrays.IntervalArray`}, default "infer"
    If "infer" the return type will be the same as *interval_array*.
    If supplied, must be done so as a keyword argument.
"""


template_doc = """
Performs a set {operation} operation.

What is considered a set is determined by the number of positional arguments used, that is, determined by the
size of *interval_arrays*.

If *interval_arrays* is empty then the sets are considered to be the intervals contained in *interval_array*.

If *interval_arrays* is not empty then the sets are considered to be the elements in *interval_arrays*, in addition to the
interval array object the accessor belongs to (an instance of :class:`pandas.IntervalIndex`, :class:`pandas.arrays.IntervalArray`).
Each of these arrays is assumed to contain disjoint intervals (and satisfy the definition of a set).  Any array containing
overlaps between intervals will be mapped to one with disjoint intervals via a union operation.

{extra_desc}
Parameters
----------
{params}

Returns
----------
:class:`pandas.IntervalIndex` or :class:`pandas.arrays.IntervalArray`

{examples}
"""

doc_difference_template = """
Performs a set difference operation.

The array elements of *interval_arrays*, and the interval array object the accessor belongs to
(an instance of :class:`pandas.IntervalIndex`, :class:`pandas.arrays.IntervalArray`) are considered to be the sets over which
the operation is performed.  Each of these arrays is assumed to contain disjoint intervals (and satisfy the definition of a set).
Any array containing overlaps between intervals will be mapped to one with disjoint intervals via a union operation.

The list *interval_arrays* must contain at least one element.  If *interval_arrays* contains a single element then
the result is the set difference between the interval array the accessor belongs to, and this single element.
If *interval_arrays* contains multiple elements then the result is the set difference between the interval array the accessor belongs to
and the union of the sets in *interval_arrays*.  This is equivalent to iteratively applying a set difference operation with each array
in *interval_arrays* as the second operand.

Each of these array operands is assumed to contain disjoint intervals (and satisfy the definition of a set).  Any array containing
overlaps between intervals will be mapped to one with disjoint intervals via a union operation.

{extra_desc}
Parameters
----------
{params}

Returns
----------
:class:`pandas.IntervalIndex` or :class:`pandas.arrays.IntervalArray`

{examples}
"""


union_params = join_params(
    [
        param_optional_args,
        param_squeeze,
        param_return_type,
    ]
)
union_docstring = template_doc.format(
    operation="union",
    extra_desc="",
    params=union_params,
    examples=union_examples,
)

intersection_params = join_params(
    [
        param_optional_args,
        param_min_overlaps,
        param_squeeze,
        param_return_type,
    ]
)
intersection_docstring = template_doc.format(
    operation="intersection",
    extra_desc="",
    params=intersection_params,
    examples=intersection_examples,
)

difference_params = join_params(
    [
        param_optional_args_difference,
        param_squeeze,
        param_return_type,
    ]
)
difference_docstring = doc_difference_template.format(
    operation="difference",
    extra_desc="",
    params=difference_params,
    examples=difference_examples,
)


symmetric_difference_params = join_params(
    [
        param_optional_args,
        param_min_overlaps,
        param_squeeze,
        param_return_type,
    ]
)
symmetric_difference_extra_desc = """
The symmetric difference can be defined as the set difference, of the union and the intersection.
The parameter *min_overlaps* in :meth:`piso.intersection`, which defines the minimum number of intervals
in an overlap required to constitute an intersection, follows through to symmetric difference under this definition.
"""
symmetric_difference_docstring = template_doc.format(
    operation="symmetric difference",
    extra_desc=symmetric_difference_extra_desc,
    params=symmetric_difference_params,
    examples=symmetric_difference_examples,
)



import pandas as pd

import piso.docstrings.interval as docstrings
from piso._decorators import Appender
from piso._exceptions import (
    ClosedMismatchError,
    ClosedValueError,
    DegenerateIntervalError,
)


def _validate_args(interval1, interval2):
    for obj in (interval1, interval2):
        if obj.length == 0:
            raise DegenerateIntervalError(obj)
        if obj.closed not in {"left", "right"}:
            raise ClosedValueError(obj)
    if interval1.closed != interval2.closed:
        raise ClosedMismatchError


@Appender(docstrings.union_docstring, join="\n", indents=1)
def union(interval1, interval2, squeeze=True):
    _validate_args(interval1, interval2)
    l1, r1 = interval1.left, interval1.right
    l2, r2 = interval2.left, interval2.right

    if r1 < l2:
        result = pd.arrays.IntervalArray([interval1, interval2])
    elif r2 < l1:
        result = pd.arrays.IntervalArray([interval2, interval1])
    else:
        result = pd.Interval(min(l1, l2), max(r1, r2), closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    return result


@Appender(docstrings.intersection_docstring, join="\n", indents=1)
def intersection(interval1, interval2, squeeze=True):
    _validate_args(interval1, interval2)
    l1, r1 = interval1.left, interval1.right
    l2, r2 = interval2.left, interval2.right

    if r1 <= l2 or r2 <= l1:
        result = pd.arrays.IntervalArray([], closed=interval1.closed)
    else:
        result = pd.Interval(max(l1, l2), min(r1, r2), closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    return result


@Appender(docstrings.difference_docstring, join="\n", indents=1)
def difference(interval1, interval2, squeeze=True):
    _validate_args(interval1, interval2)
    l1, r1 = interval1.left, interval1.right
    l2, r2 = interval2.left, interval2.right

    if r1 <= l2 or r2 <= l1:
        result = interval1
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    elif l2 <= l1 < r1 <= r2:
        result = pd.arrays.IntervalArray([], closed=interval1.closed)
    elif l1 < l2 < r2 < r1:
        result = pd.arrays.IntervalArray.from_tuples(
            [(l1, l2), (r2, r1)],
            closed=interval1.closed,
        )
    elif l1 < l2:
        result = pd.Interval(l1, l2, closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    else:  # r2 < l1
        result = pd.Interval(r2, r1, closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    return result


@Appender(docstrings.symmetric_difference_docstring, join="\n", indents=1)
def symmetric_difference(interval1, interval2, squeeze=True):
    _validate_args(interval1, interval2)
    l1, r1 = interval1.left, interval1.right
    l2, r2 = interval2.left, interval2.right

    if r1 < l2:  # separated
        result = pd.arrays.IntervalArray([interval1, interval2])
    elif r2 < l1:  # separated
        result = pd.arrays.IntervalArray([interval2, interval1])
    elif r1 == l2 or r2 == l1:  # adjacent
        result = pd.Interval(min(l1, l2), max(r1, r2), closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    elif l1 == l2 and r1 == r2:
        result = pd.arrays.IntervalArray([], closed=interval1.closed)
    elif l1 == l2:
        result = pd.Interval(min(r1, r2), max(r1, r2), closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    elif r1 == r2:
        result = pd.Interval(min(l1, l2), max(l1, l2), closed=interval1.closed)
        if not squeeze:
            result = pd.arrays.IntervalArray([result])
    else:
        result = pd.arrays.IntervalArray.from_tuples(
            [(min(l1, l2), max(l1, l2)), (min(r1, r2), max(r1, r2))],
            closed=interval1.closed,
        )
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['piso', 'piso.docstrings']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1,<2', 'staircase>=2.0.3,<3.0.0']

setup_kwargs = {
    'name': 'piso',
    'version': '0.1.0',
    'description': "Pandas Interval Set Operations: methods for set operations for pandas' Interval, IntervalArray and IntervalIndex",
    'long_description': '<p align="center"><a href="https://github.com/staircase-dev/piso"><img src="https://github.com/staircase-dev/piso/blob/master/docs/img/piso_transparent.svg" title="piso logo" alt="piso logo"></a></p>\n\n\n<p align="center">\n    <a href="https://www.python.org/" alt="Python version">\n        <img src="https://img.shields.io/pypi/pyversions/piso" /></a>\n    <a href="https://pypi.org/project/piso/" alt="PyPI version">\n        <img src="https://img.shields.io/pypi/v/piso" /></a>\n    <a href="https://anaconda.org/conda-forge/piso" alt="Conda Forge version">\n        <img src="https://anaconda.org/conda-forge/piso/badges/version.svg?branch=master&kill_cache=1" /></a>\n\t<a href="https://github.com/staircase-dev/piso/actions/workflows/ci.yml" alt="Github CI">\n\t\t<img src="https://github.com/staircase-dev/piso/actions/workflows/ci.yml/badge.svg"/></a>\n    <a href="https://piso.readthedocs.io" alt="Documentation">\n        <img src="https://readthedocs.org/projects/piso/badge/?version=latest" /></a>\n</p>\n\n\n\n# piso - pandas interval set operations\n\n**piso** exists to bring set operations (union, intersection, difference + more) to [pandas\'](https://pandas.pydata.org/) interval classes, specifically\n\n    - pandas.Interval\n    - pandas.arrays.IntervalArray\n    - pandas.IntervalIndex\n\nCurrently, there is a lack of such functionality in pandas, although it has been earmarked for development.  Until this eventuates, piso aims to fill the void.  Many of the methods can be used via accessors, which can be registered to `pandas.arrays.IntervalArray` and `pandas.IntervalIndex` classes, for example:\n\n```python\n>>> import pandas as pd\n>>> import piso\n>>> piso.register_accessors()\n\n>>> arr = pd.arrays.IntervalArray.from_tuples(\n...        [(1,5), (3,6), (2,4)]\n...    )\n\n>>> arr.piso.intersection()\n<IntervalArray>\n[(3, 4]]\nLength: 1, closed: right, dtype: interval[int64]\n```\n\nThe domain of the intervals can be either numerical, `pandas.Timestamp` or `pandas.Timedelta`.\n\nA small [case study](https://piso.readthedocs.io/en/latest/user_guide/calendar.html) using piso can be found in the [user guide](https://piso.readthedocs.io/en/latest/user_guide/index.html).  Further examples, and a detailed explanation of functionality, are provided in the [API reference](https://piso.readthedocs.io/en/latest/reference/index.html).\n\nVisit [https://piso.readthedocs.io](https://piso.readthedocs.io/) for the documentation.\n\n\n## Installation\n\n`piso` can be installed from PyPI or Anaconda.\n\nTo install the latest version from PyPI::\n\n```sh\npython -m pip install piso\n```\n\nTo install the latest version through conda-forge::\n\n```sh\nconda install -c conda-forge piso`\n```\n\n\n## Versioning\n\n[SemVer](http://semver.org/) is used by piso for versioning releases.  For versions available, see the [tags on this repository](https://github.com/staircase-dev/piso/tags).\n\n\n## License\n\nThis project is licensed under the [MIT License](https://github.com/staircase-dev/piso/blob/master/LICENSE)\n\n\n## Acknowledgments\n\nCurrently, piso is a pure-python implentation which relies heavily on [staircase](https://www.staircase.dev) and [pandas](https://pandas.pydata.org/).  It is clearly designed to operate as part of the *pandas ecosystem*.  The colours for the piso logo have been assimilated from pandas as a homage, and is not to intended to imply and affiliation with, or endorsement by, pandas.\n\nAdditionally, two classes have been borrowed, almost verbatim, from the pandas source code:\n\n    - `pandas.util._decorators.Appender`\n    - `pandas.core.accessor.CachedAccessor`',
    'author': 'Riley Clement',
    'author_email': 'venaturum@gmail.com',
    'maintainer': 'Riley Clement',
    'maintainer_email': 'venaturum@gmail.com',
    'url': 'https://github.com/staircase-dev/piso',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



class ClosedMismatchError(ValueError):
    def __init__(self):
        """
        Create a `ClosedMismatchError` indicating that interval objects did not have same `closed` values.
        """
        super().__init__(
            "The arguments must have the same value for the `closed` attribute"
        )


class ClosedValueError(ValueError):
    def __init__(self, param):
        """
        Create a `ClosedValueError` indicating that interval object did not have valid `closed` value.
        """
        super().__init__(f"Unsupported value for `closed` attribute: {param.closed}")


class DegenerateIntervalError(ValueError):
    def __init__(self, param):
        """
        Create a `DegenerateIntervalError` indicating that interval has zero length.
        """
        super().__init__(f"Zero lengths intervals: {param} - not supported by piso.")
